c*********************************************************************
c 
c File           defl.f
c Author         Jelena Ivanisevic
c Date           October, 1996
c
c Description    Polynomial deflation
c                Backward, Forward, Composite
c Reference      [1] Peters & Wilkinson paper "Problems ..."
c
c*********************************************************************
c     Forward deflation
 
      subroutine fdeflation(r,a,p)
         parameter (n=3)
         real a(0:3),p(-1:2)
         real r
 
         p(n-1) = a(n)
         do i=n-1,0,-1
            p(i-1)=a(i)+r*p(i)
         enddo
         return
      end
c
c     Backward deflation
c      
      subroutine bdeflation(r,a,q)
         parameter (n=3)
         real a(0:3),q(0:3)
         real r
 
         q(0) = - a(0)/r
         do i=1,n-1,1
            q(i)=(q(i-1)-a(i))/r
         enddo
         return
      end
 
 
c********************************************************************
c     Composite deflation
c
c     Returns
c        iflagdf  -1 if exception (NaN or infinity) occured
c                  0 deflation done but not composite
c                  1 composite deflation done
c     
c********************************************************************
 
      subroutine deflation( root, a, s, iflag )
 
      real a(0:3)
      real p(-1:2), q(0:3)
      real s(0:2)         
      integer iflag, iflagmax
      integer  k
 
      iflag = -1
 
c     If root is 0.
 
      if( root .eq. 0.) then
         do i=0,2,1    
            s(i) = a(i+1)
         enddo
         iflag = 0
*	 print *, root, s(0), s(1), s(2)
         return 
      endif
 
c     Calcuate  p from forward deflation
 
      call fdeflation(root,a,p)
        
c     Calculate q from backward deflation
 
      call bdeflation(root,a,q)
 
c     Check if any NaNs or Inf occured in coefficients
 
      do i=0,2,1    
         if ( isNaN(p(i)) .or. isInf(p(i)) 
     +        .or. isNaN(q(i)) .or. isInf(q(i))) then 
 
c           No deflation possible 
 
            iflag = -1 
            return 
 
         endif
      enddo
 
c     Calculate composite deflation coefficients
 
      call maxar( a, root, k, iflagmax )
 
c     Check if composite deflation possible
 
      if ( iflagmax .eq. -1 ) then
 
c        We can not guarantee specification since composite deflation  
c        not performed accurately
 
         iflag = 0
 
      else 
 
c        Composite deflation successful
 
         iflag = 1
 
      endif
 
c     Combine backward and forward deflation based on k
 
      do i=0,2,1
            if (i.lt.k) then
               s(i) = q(i)
            else 
               s(i) = p(i)
            endif
      enddo
*      print *, root, s(0), s(1), s(2)

 
      end 
 
c********************************************************************
c
c     Calculates k where |a(k)*r^k| is maximal for k= 2,0
c
c     This is described in [1] except that is not quite clear 
c     what to do if a(3)*r*r*r is maximum i.e. k = 3
c     I think that it should be complete backward deflation but 
c     for the purpose of this program I have used k = 2 for 
c     this special case because I have already derived my specs
c     based on this assumption.
 
c     This part of defaltion code does not check if NaNs occured.
c     it should be notifed to the qubic routine in that case that
c     composit deflation has not been successful and spec not 
c     guaranteed...
c      
c     Returns
c        iflag 
c            0  success
c           -1 NaN or Infinity occured during calculation
c
c********************************************************************
   
      subroutine maxar( a, root, k, iflag )   
         real a(0:3)
         real root
         integer k
         real tmax,rmax
         integer iflag 
         
         iflag = -1
         k = 0
         rmax = abs(a(0))
         tmax = abs(a(1)*root)
 
c        Is ( a(1) * root ) max?         
        
         if (tmax .gt. rmax) then 
            rmax = tmax 
            k = 1
         endif
 
c        Is ( a(2) * root^2 ) max?         
 
         if( isInf(tmax) ) return
         tmax = abs(a(2)*root*root)
 
         if (tmax .gt. rmax) then 
            rmax = tmax 
            k = 2
         endif
 
c        Is ( a(3) * root^3 ) max?
 
         if( isInf(tmax) .or. isNaN(tmax) ) return
         tmax = abs(a(3)*root*root*root)
        
         if (tmax .gt. rmax) then 
            rmax = tmax 
            k = 2
         endif
 
         iflag = 0
 
      return 
      end    
